@@ -160,7 +160,7 @@ main (int   argc,
   g_test_init (&argc, &argv, NULL);
 
   g_test_add_func ("/builder/complex", test_builder_complex);
-  g_test_add_func ("/builder/complex", test_builder_empty);
+  g_test_add_func ("/builder/empty", test_builder_empty);
   g_test_add_func ("/builder/reset", test_builder_reset);
 
   return g_test_run ();
